from dendrotweaks.morphology.reduce.reduce import map_segs_to_params
from dendrotweaks.morphology.reduce.reduce import map_segs_to_locs
from dendrotweaks.morphology.reduce.reduce import map_segs_to_reduced_segs
from dendrotweaks.morphology.reduce.reduce import map_reduced_segs_to_params

from dendrotweaks.morphology.reduce.reduce import set_avg_params_to_reduced_segs
from dendrotweaks.morphology.reduce.reduce import interpolate_missing_values


import neuron_reduce
from dendrotweaks.morphology.reduce.reduced_cylinder import _get_subtree_biophysical_properties
neuron_reduce.reducing_methods._get_subtree_biophysical_properties = _get_subtree_biophysical_properties
from neuron_reduce.reducing_methods import reduce_subtree as get_unique_cable_properties

from dendrotweaks.morphology.reduce.reduced_cylinder import calculate_nsegs
from dendrotweaks.morphology.reduce.reduced_cylinder import apply_params_to_section
